# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mlflow_artifacts.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import service as _service
from google.protobuf import service_reflection
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from .scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2
from . import databricks_pb2 as databricks__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='mlflow_artifacts.proto',
  package='mlflow.artifacts',
  syntax='proto2',
  serialized_options=_b('\n\024org.mlflow.api.proto\220\001\001\342?\002\020\001'),
  serialized_pb=_b('\n\x16mlflow_artifacts.proto\x12\x10mlflow.artifacts\x1a\x15scalapb/scalapb.proto\x1a\x10\x64\x61tabricks.proto\"\x1e\n\x10\x44ownloadArtifact\x1a\n\n\x08Response\"\x1c\n\x0eUploadArtifact\x1a\n\n\x08Response\"T\n\rListArtifacts\x12\x0c\n\x04path\x18\x01 \x01(\t\x1a\x35\n\x08Response\x12)\n\x05\x66iles\x18\x01 \x03(\x0b\x32\x1a.mlflow.artifacts.FileInfo\";\n\x08\x46ileInfo\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x0e\n\x06is_dir\x18\x02 \x01(\x08\x12\x11\n\tfile_size\x18\x03 \x01(\x03\x32\xaf\x04\n\x16MlflowArtifactsService\x12\xbd\x01\n\x10\x64ownloadArtifact\x12\".mlflow.artifacts.DownloadArtifact\x1a+.mlflow.artifacts.DownloadArtifact.Response\"X\xf2\x86\x19T\n=\n\x03GET\x12\x30/mlflow-artifacts/artifacts/<path:artifact_path>\x1a\x04\x08\x02\x10\x00\x10\x01*\x11\x44ownload Artifact\x12\xb5\x01\n\x0euploadArtifact\x12 .mlflow.artifacts.UploadArtifact\x1a).mlflow.artifacts.UploadArtifact.Response\"V\xf2\x86\x19R\n=\n\x03PUT\x12\x30/mlflow-artifacts/artifacts/<path:artifact_path>\x1a\x04\x08\x02\x10\x00\x10\x01*\x0fUpload Artifact\x12\x9c\x01\n\rlistArtifacts\x12\x1f.mlflow.artifacts.ListArtifacts\x1a(.mlflow.artifacts.ListArtifacts.Response\"@\xf2\x86\x19<\n(\n\x03GET\x12\x1b/mlflow-artifacts/artifacts\x1a\x04\x08\x02\x10\x00\x10\x01*\x0eList ArtifactsB\x1e\n\x14org.mlflow.api.proto\x90\x01\x01\xe2?\x02\x10\x01')
  ,
  dependencies=[scalapb_dot_scalapb__pb2.DESCRIPTOR,databricks__pb2.DESCRIPTOR,])




_DOWNLOADARTIFACT_RESPONSE = _descriptor.Descriptor(
  name='Response',
  full_name='mlflow.artifacts.DownloadArtifact.Response',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=105,
  serialized_end=115,
)

_DOWNLOADARTIFACT = _descriptor.Descriptor(
  name='DownloadArtifact',
  full_name='mlflow.artifacts.DownloadArtifact',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_DOWNLOADARTIFACT_RESPONSE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=85,
  serialized_end=115,
)


_UPLOADARTIFACT_RESPONSE = _descriptor.Descriptor(
  name='Response',
  full_name='mlflow.artifacts.UploadArtifact.Response',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=105,
  serialized_end=115,
)

_UPLOADARTIFACT = _descriptor.Descriptor(
  name='UploadArtifact',
  full_name='mlflow.artifacts.UploadArtifact',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_UPLOADARTIFACT_RESPONSE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=117,
  serialized_end=145,
)


_LISTARTIFACTS_RESPONSE = _descriptor.Descriptor(
  name='Response',
  full_name='mlflow.artifacts.ListArtifacts.Response',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='files', full_name='mlflow.artifacts.ListArtifacts.Response.files', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=178,
  serialized_end=231,
)

_LISTARTIFACTS = _descriptor.Descriptor(
  name='ListArtifacts',
  full_name='mlflow.artifacts.ListArtifacts',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='path', full_name='mlflow.artifacts.ListArtifacts.path', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_LISTARTIFACTS_RESPONSE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=147,
  serialized_end=231,
)


_FILEINFO = _descriptor.Descriptor(
  name='FileInfo',
  full_name='mlflow.artifacts.FileInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='path', full_name='mlflow.artifacts.FileInfo.path', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='is_dir', full_name='mlflow.artifacts.FileInfo.is_dir', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_size', full_name='mlflow.artifacts.FileInfo.file_size', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=233,
  serialized_end=292,
)

_DOWNLOADARTIFACT_RESPONSE.containing_type = _DOWNLOADARTIFACT
_UPLOADARTIFACT_RESPONSE.containing_type = _UPLOADARTIFACT
_LISTARTIFACTS_RESPONSE.fields_by_name['files'].message_type = _FILEINFO
_LISTARTIFACTS_RESPONSE.containing_type = _LISTARTIFACTS
DESCRIPTOR.message_types_by_name['DownloadArtifact'] = _DOWNLOADARTIFACT
DESCRIPTOR.message_types_by_name['UploadArtifact'] = _UPLOADARTIFACT
DESCRIPTOR.message_types_by_name['ListArtifacts'] = _LISTARTIFACTS
DESCRIPTOR.message_types_by_name['FileInfo'] = _FILEINFO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DownloadArtifact = _reflection.GeneratedProtocolMessageType('DownloadArtifact', (_message.Message,), dict(

  Response = _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), dict(
    DESCRIPTOR = _DOWNLOADARTIFACT_RESPONSE,
    __module__ = 'mlflow_artifacts_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.artifacts.DownloadArtifact.Response)
    ))
  ,
  DESCRIPTOR = _DOWNLOADARTIFACT,
  __module__ = 'mlflow_artifacts_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.artifacts.DownloadArtifact)
  ))
_sym_db.RegisterMessage(DownloadArtifact)
_sym_db.RegisterMessage(DownloadArtifact.Response)

UploadArtifact = _reflection.GeneratedProtocolMessageType('UploadArtifact', (_message.Message,), dict(

  Response = _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), dict(
    DESCRIPTOR = _UPLOADARTIFACT_RESPONSE,
    __module__ = 'mlflow_artifacts_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.artifacts.UploadArtifact.Response)
    ))
  ,
  DESCRIPTOR = _UPLOADARTIFACT,
  __module__ = 'mlflow_artifacts_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.artifacts.UploadArtifact)
  ))
_sym_db.RegisterMessage(UploadArtifact)
_sym_db.RegisterMessage(UploadArtifact.Response)

ListArtifacts = _reflection.GeneratedProtocolMessageType('ListArtifacts', (_message.Message,), dict(

  Response = _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), dict(
    DESCRIPTOR = _LISTARTIFACTS_RESPONSE,
    __module__ = 'mlflow_artifacts_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.artifacts.ListArtifacts.Response)
    ))
  ,
  DESCRIPTOR = _LISTARTIFACTS,
  __module__ = 'mlflow_artifacts_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.artifacts.ListArtifacts)
  ))
_sym_db.RegisterMessage(ListArtifacts)
_sym_db.RegisterMessage(ListArtifacts.Response)

FileInfo = _reflection.GeneratedProtocolMessageType('FileInfo', (_message.Message,), dict(
  DESCRIPTOR = _FILEINFO,
  __module__ = 'mlflow_artifacts_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.artifacts.FileInfo)
  ))
_sym_db.RegisterMessage(FileInfo)


DESCRIPTOR._options = None

_MLFLOWARTIFACTSSERVICE = _descriptor.ServiceDescriptor(
  name='MlflowArtifactsService',
  full_name='mlflow.artifacts.MlflowArtifactsService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=295,
  serialized_end=854,
  methods=[
  _descriptor.MethodDescriptor(
    name='downloadArtifact',
    full_name='mlflow.artifacts.MlflowArtifactsService.downloadArtifact',
    index=0,
    containing_service=None,
    input_type=_DOWNLOADARTIFACT,
    output_type=_DOWNLOADARTIFACT_RESPONSE,
    serialized_options=_b('\362\206\031T\n=\n\003GET\0220/mlflow-artifacts/artifacts/<path:artifact_path>\032\004\010\002\020\000\020\001*\021Download Artifact'),
  ),
  _descriptor.MethodDescriptor(
    name='uploadArtifact',
    full_name='mlflow.artifacts.MlflowArtifactsService.uploadArtifact',
    index=1,
    containing_service=None,
    input_type=_UPLOADARTIFACT,
    output_type=_UPLOADARTIFACT_RESPONSE,
    serialized_options=_b('\362\206\031R\n=\n\003PUT\0220/mlflow-artifacts/artifacts/<path:artifact_path>\032\004\010\002\020\000\020\001*\017Upload Artifact'),
  ),
  _descriptor.MethodDescriptor(
    name='listArtifacts',
    full_name='mlflow.artifacts.MlflowArtifactsService.listArtifacts',
    index=2,
    containing_service=None,
    input_type=_LISTARTIFACTS,
    output_type=_LISTARTIFACTS_RESPONSE,
    serialized_options=_b('\362\206\031<\n(\n\003GET\022\033/mlflow-artifacts/artifacts\032\004\010\002\020\000\020\001*\016List Artifacts'),
  ),
])
_sym_db.RegisterServiceDescriptor(_MLFLOWARTIFACTSSERVICE)

DESCRIPTOR.services_by_name['MlflowArtifactsService'] = _MLFLOWARTIFACTSSERVICE

MlflowArtifactsService = service_reflection.GeneratedServiceType('MlflowArtifactsService', (_service.Service,), dict(
  DESCRIPTOR = _MLFLOWARTIFACTSSERVICE,
  __module__ = 'mlflow_artifacts_pb2'
  ))

MlflowArtifactsService_Stub = service_reflection.GeneratedServiceStubType('MlflowArtifactsService_Stub', (MlflowArtifactsService,), dict(
  DESCRIPTOR = _MLFLOWARTIFACTSSERVICE,
  __module__ = 'mlflow_artifacts_pb2'
  ))


# @@protoc_insertion_point(module_scope)
